/* broker.c
 * Copyright (c) 2000 by Not a Number B.V. (NaN) and Hans Lambermont.
 * This file is covered by the license in the accompanying "LICENSE" file.
 * $Id: broker.c 87 2004-10-27 20:43:35Z goofster $
 */

// system includes
#include <string.h>
#include <stdio.h>
#include "plgetopt.h"
#include "nspr.h"
#include <signal.h>

// our own includes
#include "nspr-additions.h"
#include "compat.h"
#include "scan.h"
#include "common.h"
#include "blendfile.h"
#include "parser.h"
#include "broker.h"


int main(int argc, char **argv) {
	PRFileDesc *clientSock;
	PRNetAddr listenAddr, clientAddr;
	void sig_int(int);
	int i;
	PLOptStatus os;
	PLOptState *opt;

	PR_STDIO_INIT();

	// Print program name and version
	fprintf(stdout, "Blender Renderdaemon Broker\n");
	fprintf(stdout, "Version: %s\n\n", version);	

	// Read configuration information from a file if one exists
	configInfo = parse_config();	

	// Set debug level from config file
	debug = configInfo->debug;

	// Read and execute any command line arguments
	opt = PL_CreateOptState(argc, argv, "n:x:i:o:p:m:a:d:h");

	while (PL_OPT_EOL != (os = PL_GetNextOpt(opt)))
	{
		if (PL_OPT_BAD == os) continue;

		switch (opt->option)
		{
			case 0:
				if (debug>1)
				{
					fprintf(stdout, "option %c", opt->option);
					if (opt->value)
						fprintf(stdout, " with arg %s", opt->value);
					fprintf(stdout, "\n");
				}
				break;
			case 'a':
				if (debug>1) fprintf(stdout, "maxport set to %s\n", opt->value);
				configInfo->maxPort = atoi(opt->value);
				break;
			case 'd':
				if (debug>1) fprintf(stdout, "debug set to %s\n", opt->value);
				debug = atoi(opt->value);
				break;
			case 'h':
				print_help();
				exit(0);
			case 'i':
				if (debug>1) fprintf(stdout, "input directory set to %s\n", opt->value);
				free(configInfo->brokertasks);
				configInfo->brokertasks = strdup(opt->value);
				break;
			case 'm':
				if (debug>1) fprintf(stdout, "minport set to %s\n", opt->value);
				configInfo->minPort = atoi(opt->value);
				break;
			case 'n':
				if (debug>1) fprintf(stdout, "minsize set to %s\n", opt->value);
				configInfo->minBlendFileSize = atol(opt->value);
				break;
			case 'o':
				if (debug>1) fprintf(stdout, "done directory set to %s\n", opt->value);
				free(configInfo->brokerdone);
				configInfo->brokerdone = strdup(opt->value);
				break;
			case 'p':
				if (debug>1) fprintf(stdout, "broker port set to %s\n", opt->value);
				configInfo->brokerPort = atoi(opt->value);
				break;
			case 'x':
				if (debug>1) fprintf(stdout, "maxsize set to %s\n", opt->value);
				configInfo->maxBlendFileSize = atol(opt->value);
				break;
			default:
				print_help();
				exit(0);
		}
	}

	PL_DestroyOptState(opt);

	if(debug>1) signal(SIGINT, blender_esc);

	// This mutex controls broker to renderd connections
	BlenderActionMutex = PR_NewLock();
	BlenderActionCondition = PR_NewCondVar(BlenderActionMutex);

	SingleShotMutex = PR_NewLock();
	SingleShotCondition = PR_NewCondVar(SingleShotMutex);

	InfoMutex = PR_NewLock();
	mainThreadMutex = PR_NewLock();

	PR_Lock(InfoMutex);
	for (i=0; i<MAXTHREADS; i++) {
		Info[i].occupied = FALSE;
		Info[i].frames_rendered = 0;
	}
	PR_Unlock(InfoMutex);

	PR_Lock(mainThreadMutex);
	mainThread.renderThreadfd = 0;
	mainThread.frames_rendered = 0;
	mainThread.epoch = PR_Now();
	mainThread.msecPerFrame = 0;
	mainThread.blendFilePolling = TRUE;
	mainThread.status = MT_STATUS_INITIALIZE;

	for (i=0; i<MAXTHREADS; i++) {
		mainThread.nsprfd[i] = 0;
	}
	PR_Unlock(mainThreadMutex);

	port.local = configInfo->brokerPort;
	port.remote = configInfo->renderdPort;
	brokerScanDir = configInfo->brokertasks;
	brokerWorkDone = B_TASK;

	listenSock = PR_NewTCPSocket();
	memset(&listenAddr, 0 , sizeof(listenAddr));
	listenAddr.inet.family = PR_AF_INET;
	listenAddr.inet.port = PR_htons(port.local);
	listenAddr.inet.ip = PR_htonl(PR_INADDR_ANY);
	if (PR_Bind(listenSock, &listenAddr) == PR_FAILURE)
	{
		if (debug>0) fprintf(stderr, "Could not create socket. Error %s (just wait TIME_WAIT)\n", PRB_GetErrorText());
		exit(1);
	}
	if (PR_Listen(listenSock, 10) == PR_FAILURE)
	{
		if (debug>0) fprintf(stderr, "unable to listen\n");
		exit(1);
	}
	// this shouldn't be here ...
	BlenderAction = BA_NOACTION;
	SingleShot.start = 1;
	SingleShot.end = 0; 	// init to 0 so status makes sense

	// start control thread, also used for blendFilePolling
	if ( ! PR_CreateThread(PR_USER_THREAD, broker_control, NULL, PR_PRIORITY_NORMAL, PR_GLOBAL_THREAD, PR_UNJOINABLE_THREAD, 0))
	{
		if (debug>0) fprintf(stderr, "Could not create dumb thread %s (%d)!\n", PRB_GetErrorText(), PR_GetOSError());
		exit(1);
	}

	// main thread starting loop
	for ( ; ; ) {
		if (mainThread.status == MT_STATUS_INITIALIZE) {
			if (debug>0) fprintf(stdout, "Initializing Broker...\n");
		}
		if (debug>0) fprintf(stdout, "Ready\n\n");
		if ((clientSock = PR_Accept(listenSock, &clientAddr, FOREVER)) != NULL) {
			// Save clientAddr info, indexed by socket file descriptor
			struct threadArgsType *threadArgs =
			malloc(sizeof(struct threadArgsType));
			// use the real filedescriptor, good for debugging
			int fd = PRB_GetFileDesc(clientSock);
			if (fd < MAXTHREADS) {
				threadArgs->fd = fd;
				threadArgs->sock = clientSock;
				threadArgs->addr = clientAddr;
				mainThread.status = MT_STATUS_IDLE;
				if ( ! PR_CreateThread(PR_USER_THREAD, broker_thread, threadArgs, PR_PRIORITY_NORMAL, PR_GLOBAL_THREAD, PR_UNJOINABLE_THREAD, 0))
				{
					if (debug>1) fprintf(stderr, "Could not create dumb thread %s (%d)!\n", PRB_GetErrorText(), PR_GetOSError());
				}
			} else {
				PRB_Close(clientSock);
				if (debug>1) fprintf(stderr, "Too many threads (max=%i)\n", MAXTHREADS-1);
		    }
		} else if (StopDaemon) break;
	}
	if (debug>1) fprintf(stdout, "Main thread exit\n");
	return(0);
}

/* handle blendFilePolling
   handle frame rendering statistics
   redistribute failed frames for broken connections
 */
void PR_CALLBACK broker_control(void *arg) {
	int blendFilePolling = FALSE, OldblendFilePolling = FALSE;
	int FirstTime = FALSE, status = FALSE;

	//	  PR_Sleep(PR_SecondsToInterval(10)); Why insert 10 seconds of nothing here?
	for ( ; ; ) {
		PR_Lock(mainThreadMutex);
		blendFilePolling = mainThread.blendFilePolling;
		status = mainThread.status;
		PR_Unlock(mainThreadMutex);
		// Make blendFilePolling changes visible
		FirstTime = FALSE;
		if (mainThread.status != MT_STATUS_INITIALIZE) {
			if (blendFilePolling == TRUE) {
				if (OldblendFilePolling == FALSE) {
					if (debug>1) fprintf(stdout, "broker_control: blendFilePolling = TRUE\n");
					FirstTime = TRUE;
				}
			} else {
				if (OldblendFilePolling == TRUE) {
					if (debug>1) fprintf(stdout, "broker_control: blendFilePolling = FALSE\n");
					FirstTime = TRUE;
				}
			}
			OldblendFilePolling = blendFilePolling;
			if (blendFilePolling == TRUE) {
				if (status == MT_STATUS_IDLE) {
					// Scan brokertasks directory for blend files.
				 	brokerScan(configInfo->brokertasks);
				}
			}
			// Copy blend file to work directory
			brokerFileCopy (configInfo->brokertasks, configInfo->brokerwork);
		}
		// this is the interval between checks for new files
		PR_Sleep(PR_SecondsToInterval(10));
	}
}

void set_blendFilePolling(char *line) {
	if (Bstrncasecmp(line,"on", 2) == 0) {
		PR_Lock(mainThreadMutex);
		mainThread.blendFilePolling = TRUE;
		PR_Unlock(mainThreadMutex);
	} else {
		PR_Lock(mainThreadMutex);
		mainThread.blendFilePolling = FALSE;
		PR_Unlock(mainThreadMutex);
	}
}

void PR_CALLBACK broker_thread(void *arg) {
	struct threadArgsType *args = (struct threadArgsType*) arg;
	PRFileDesc *sockfd = args->sock;
	int fd = args->fd;
	PRInt32 nread;
	char line[MAXLINE];

	PRB_Msg(sockfd, "220 Blender render broker %s ready", version);
	for ( ; ; ) {
		if ( (nread = PRB_ReadLine(sockfd, line, MAXLINE)) == 0)
			break;		// connection closed by other end
		if (nread<0) {
			if (debug>1) fprintf(stdout, "broker thread exits\n");
			break;
		}
		if (debug>1) fprintf(stdout, "%i< %s\n", fd, line);
		if (Bstrncasecmp(line,"Help",4) == 0) {
			showhelp(sockfd);
		}else if (Bstrncasecmp(line,"Quit",4) == 0) {
			PRB_Msg(sockfd, "210 Broker connection will be closed");
			PRB_Close(sockfd);
			break;
		} else if (Bstrncasecmp(line,"Status",6) == 0) {
			showstatus(args);
		} else if (Bstrncasecmp(line,"Config",6) == 0) {
			showconfig(sockfd);
		} else if (Bstrncasecmp(line,"Debug ", 6) == 0) {
			debug = atoi(line + 6);
			PRB_Msg(sockfd, "210 debug set to %i", debug);
		} else if (Bstrncasecmp(line,"StopAllDaemons",14) == 0) {
			stopalldaemons(sockfd);
			break;
		} else if (Bstrncasecmp(line,"StopDaemon",10) == 0) {
			stopdaemon(sockfd);
			break;
		} else if (Bstrncasecmp(line,"NewRenderd",10) == 0) {
			newrenderd(args, line + 10);
		} else if (Bstrncasecmp(line,"NewBlendFile Name ",18) == 0) {
			if (read_blenderfile(sockfd, line + 18) == 0) {
				if (debug>1) fprintf(stdout, "   read_blenderfile is OK\n");
				// file read ok, send action to threads
				SignalThreads(BA_NEWBLENDFILE);
			}
		} else if (Bstrncasecmp(line,"UploadNewBlendFile",18) == 0) {			
			if(receive_BlendFile(sockfd) == 0){
				if (debug>1) fprintf(stdout, "Error receiving file\n");
			}else{
				if (debug>1) fprintf(stdout, "File received\n");
				SignalThreads(BA_NEWBLENDFILE);
			}					
		} else if (Bstrncasecmp(line,"StartFrame ", 11) == 0) {
			SetFrameLimit(STARTLIMIT, atoi(line + 11));
		} else if (Bstrncasecmp(line,"EndFrame ", 9) == 0) {
			SetFrameLimit(ENDLIMIT, atoi(line + 9));
		} else if (Bstrncasecmp(line,"blendFilePolling ", 17) == 0) {
			set_blendFilePolling(line + 17);
		} else if (Bstrncasecmp(line,"Scan", 4) == 0) {
			if ( ! PR_CreateThread(PR_USER_THREAD, broker_scan_thread, (void *) &port, PR_PRIORITY_NORMAL, PR_GLOBAL_THREAD, PR_UNJOINABLE_THREAD, 0) ) {
				PRB_Msg(sockfd, "500 Cannot create scan thread");
			}
		}
	}
	free(args);
}

void showhelp(PRFileDesc *sockfd)
{
	PRB_Msg(sockfd, "210-This is the Blender renderd broker");
	PRB_Msg(sockfd, "210-Available commands are:");
	PRB_Msg(sockfd, "210-Help                     this helpfile");
	PRB_Msg(sockfd, "210-Quit                     end connection");
	PRB_Msg(sockfd, "210-StopDaemon               make broker exit. handle with care");
	PRB_Msg(sockfd, "210-StopAllDaemons           make broker and all connected daemons exit.");
	PRB_Msg(sockfd, "210-                         handle with great care");
	PRB_Msg(sockfd, "210-Status                   give broker status");
	PRB_Msg(sockfd, "210-Config                   show broker configuration settings");
	PRB_Msg(sockfd, "210-Debug xxx                set debugging to level xxx");
	PRB_Msg(sockfd, "210-blendFilePolling sss     set blendFilePolling 'on' or 'off'");
	PRB_Msg(sockfd, "210-NewRenderd hhh xxx       add host hhh port xxx to renderfarm");
	PRB_Msg(sockfd, "210-NewBlendFile Name nnn    use blendfile nnn");
	PRB_Msg(sockfd, "210-StartFrame xxx           set start frame xxx of current blenderfile");
	PRB_Msg(sockfd, "210-EndFrame xxx             set end frame xxx of current blenderfile");
	PRB_Msg(sockfd, "210-                         and start rendering");
	PRB_Msg(sockfd, "210-Scan                     scan for renderd clients on network");
}

void showconfig(PRFileDesc *sockfd)
{
	PRB_Msg(sockfd, "210-This is the Blender renderd broker");
	PRB_Msg(sockfd, "210-Current configuration settings:");
	PRB_Msg(sockfd, "210- Min Port:          %i",configInfo->minPort);
	PRB_Msg(sockfd, "210- Max Port:          %i",configInfo->maxPort);
	PRB_Msg(sockfd, "210- Broker Port:       %i",configInfo->brokerPort);
	PRB_Msg(sockfd, "210- Input Dir:         %s",configInfo->brokertasks);
	PRB_Msg(sockfd, "210- Work Dir:          %s" ,configInfo->brokerwork);
	PRB_Msg(sockfd, "210- Output Dir:        %s",configInfo->brokerdone);
	PRB_Msg(sockfd, "210- Min File Size:     %li",configInfo->minBlendFileSize);
	PRB_Msg(sockfd, "210- Max File Size:     %li",configInfo->maxBlendFileSize);
	PRB_Msg(sockfd, "210- Debug:             %i",debug);
}


void SetFrameLimit(int limittype, int limit) {
	PR_Lock(SingleShotMutex);
	switch (limittype) {
		case STARTLIMIT:
			SingleShot.start = limit;
			if (debug>0) fprintf(stdout, "SingleShot start = %d\n",
					  SingleShot.start);
			break;
		case ENDLIMIT:
			SingleShot.end = limit;
			if (debug>0)
			fprintf(stdout, "SingleShot start = %d end = %d. Start rendering!\n",
			    SingleShot.start, SingleShot.end);
			// ok, we know the endframe. Signal the render threads
			SingleShot.frame = SingleShot.start;
			PR_Lock(mainThreadMutex);
			mainThread.status = MT_STATUS_ACTIVE;
			brokerJob.status = B_CURRENT;
			PR_Unlock(mainThreadMutex);
			SignalThreads(BA_SINGLESHOTRENDER);
			break;
		default:
			// yup, error++ TODO
			break;
	}
	PR_Unlock(SingleShotMutex);
}

void SignalThreads(int signal) {
	PR_Lock(BlenderActionMutex);
	BlenderAction = signal;
	PR_NotifyAllCondVar(BlenderActionCondition);
	PR_Unlock(BlenderActionMutex);
}

int SingleShotLoop(struct threadArgsType *args) {
	PRFileDesc *sockfd = args->sock2;
	struct SingleShotTag myshot;
	int finished = 0;
	char FrameName[MAXLINE];
	LONGINT FrameSize;
	int framenr;
	PRInt32 nread;
	char line[MAXLINE];


	for ( ; ; ) {
		int lastone = FALSE;
		brokerWorkDone = B_WORK;
		PR_Lock(SingleShotMutex);
		// save current SingleShot struct
		myshot = SingleShot;
		if (SingleShot.frame <= SingleShot.end) {
			// increase todo
			SingleShot.frame++;
			if (SingleShot.frame >= SingleShot.end) {
				// Yay, we're doing the last one
				lastone = TRUE;
				brokerWorkDone = B_DONE;
				brokerJob.status = B_COMPLETE;
			}
		} else {
			finished = 1;
			brokerFileCopy(configInfo->brokerwork, configInfo->brokerdone);
		}
		PR_Unlock(SingleShotMutex);

		if (finished) {
			brokerWorkDone = B_TASK;
			break;
		}

		PRB_Msg(sockfd, "SingleShotRender %d", myshot.frame);

		// let blender at the renderd chew on this frame

		if ( (nread = PRB_ReadLine(sockfd, line, MAXLINE)) <= 0) {
			if (debug>0) fprintf(stderr, "%i ERROR renderd hung up\n", args->fd2);
			// TODO fix SingleShot.frame failure
			return(-1); 	// connection closed by other end
		}
		if (debug>1) fprintf(stdout, "%i Command = [%s][%d]\n", args->fd2, line, nread);
		// check errorcodes ! TODO
		if (Bstrncasecmp(line,"250 Prepare for frame ",22) == 0) {
			sscanf(line+22, "%i %s upload of %ld bytes\n", &framenr, FrameName,
			&FrameSize);
			if (debug>1) fprintf(stdout, "%d download %s of %ld bytes\n",
			args->fd2, FrameName, FrameSize);
			SaveFrame(args, FrameName, FrameSize);
		} else {
			if (debug>0) fprintf(stdout, "%d ERROR unknown response from renderd\n", args->fd2);
			// TODO fix SingleShot.frame failure
			return(-2);
		}

		// TODO Not completely right, but workable for now.
		if (lastone == TRUE) {
			PR_Lock(mainThreadMutex);
			mainThread.status = MT_STATUS_IDLE;
			PR_Unlock(mainThreadMutex);
			if (debug>1) fprintf(stdout, "Done with last frame, unlocking main thread.\n");
		}
	}
	return(0);
}

void SaveFrame(struct threadArgsType *args, char *FrameName, LONGINT FrameSize) {
	PRFileDesc *sockfd = args->sock2;
	FILE *framefd;
	void *frame;
	PRIntervalTime startTime = PR_IntervalNow();
	PRUint32 timeDiff;

	frame = malloc(FrameSize);
	if (PRB_Recv(sockfd, frame, FrameSize, 0, PR_INTERVAL_NO_TIMEOUT) == FrameSize) {
		// optionally remove directory part of FrameName
		// TODO
		if ((framefd = fopen(FrameName, "wb")) == NULL) {
			if (debug>1) fprintf(stdout, "%i ERROR cannot open for writing %s\n", args->fd2, FrameName);
			// TODO
		} else {
			if (fwrite(frame, FrameSize, 1, framefd) != 1) {
				if (debug>1) fprintf(stdout, "%i ERROR cannot write %s\n", args->fd2, FrameName);
				// TODO
			}
			fclose(framefd);
			if (debug>0) fprintf(stdout, "%i frame %s saved\n", args->fd2, FrameName);
			//Update status
			timeDiff = PR_IntervalToMilliseconds(PR_IntervalNow() - startTime);
			PR_Lock(mainThreadMutex);
			// for this fd
			++mainThread.frames_rendered;
			mainThread.msecPerFrame = (
				mainThread.msecPerFrame *
				mainThread.frames_rendered + timeDiff) /
				(mainThread.frames_rendered);
			PR_Unlock(mainThreadMutex);

			PR_Lock(InfoMutex);
			Info[args->fd].msecPerFrame = (
				Info[args->fd].msecPerFrame *
				Info[args->fd].frames_rendered + timeDiff) /
				++(Info[args->fd].frames_rendered);
			PR_Unlock(InfoMutex);
		}
	} else {
		if (debug>1) fprintf(stdout, "%i ERROR frame %s not received properly\n", args->fd2, FrameName);
		// do something useful TODO
	}
	free(frame);
}

void newrenderd(struct threadArgsType *args, char *linep) {
	PRFileDesc *sockfd = args->sock;
	struct threadArgsType *args2 = malloc(sizeof(struct threadArgsType));

	// use some default port value
	args->port = configInfo->minPort;

	sscanf(linep, "%s %i", args->server, &(args->port));

	memcpy(args2, args, sizeof(struct threadArgsType));

	if ((configInfo->minPort <= args->port) && (args->port <= configInfo->maxPort)) {
		if ( ! PR_CreateThread(PR_USER_THREAD, brokertorenderd, args2, PR_PRIORITY_NORMAL, PR_GLOBAL_THREAD, PR_UNJOINABLE_THREAD, 0)) {
			if (debug>1) fprintf(stderr, "ERROR Could not create dumb thread %s (%d)!\n", PRB_GetErrorText(), PR_GetOSError());
		}
	} else {
		PRB_Msg(sockfd, "501 Unacceptable port number");
	}
	if (brokerWorkDone == B_WORK) {
		if (debug>0) fprintf (stdout, "Checking Brokerwork directory:\n");
		brokerScan(configInfo->brokerwork);
	}
}

void PR_CALLBACK brokertorenderd(void *arg) {
	struct threadArgsType *args = malloc(sizeof(struct threadArgsType));
	char cport[MAXLINE];
	PRFileDesc *btrSock; // BrokerToRenderdaemonSocket
	PRNetAddr btrAddr;
	PRHostEnt host;
	int fd;
	char buffer[PR_NETDB_BUF_SIZE];
	char line[MAXLINE];
	PRInt32 nread;
	struct threadArgsType *args2 = (struct threadArgsType *) arg;

	memcpy(args, arg, sizeof(struct threadArgsType));

	free(args2);

	if (debug>1) fprintf(stdout, "%d ordered a new thread to connect to %s:%i\n", args->fd, args->server, args->port);

	sprintf(cport, "%i", args->port);
	if (PR_GetHostByName(args->server, buffer, PR_NETDB_BUF_SIZE, &host)==PR_FAILURE) {
		if (debug>0) fprintf(stderr, "ERROR PR_GetHostByName : %s for %s\n", PRB_GetErrorText(), args->server);
		free (args);
		return;
	}
	PR_EnumerateHostEnt(0, &host, (PRUint16) args->port, &btrAddr);
	btrSock = PR_NewTCPSocket();
	if (PR_Connect(btrSock, &btrAddr, PR_INTERVAL_NO_TIMEOUT)==PR_FAILURE) {
		if (debug>0) fprintf(stderr, "ERROR PR_Connect %s (%d)\n", PRB_GetErrorText(), PR_GetOSError());
		free (args);
		return;
	}

	// save btrSock descriptor for debugging purposes
	args->fd2 = PRB_GetFileDesc(btrSock);
	fd = args->fd2;
	args->sock2 = btrSock;
	args->addr2 = btrAddr;

	// update clientInfo
	PR_Lock(InfoMutex);
	Info[fd].occupied = TRUE;
	Info[fd].epoch = PR_Now();
	PR_NetAddrToString(&args->addr2, Info[fd].ip, 256);
	Info[fd].port = PR_ntohs(args->addr2.inet.port);
	if (debug>1) fprintf(stdout, "%i connects to %s:%i (%s:%i ordered by %i)\n", fd, Info[fd].ip, Info[fd].port, args->server, args->port, args->fd);
	PR_Unlock(InfoMutex);

	// Read renderd welcome message
	if ( (nread = PRB_ReadLine(btrSock, line, MAXLINE)) <= 0) {
		if (debug>1) fprintf(stdout, "broker to renderd thread %i exits\n", fd);
		free (args);
		return; // connection closed by other end
	}
	// todo: check for 220
	for ( ; ; ) {
		// all communication to/from renderd goes here, wait for
		// pthread_cond_broadcast
		if (debug) fprintf(stdout, "%i awaits signal\n", fd);
		PR_Lock(BlenderActionMutex);
		PR_WaitCondVar(BlenderActionCondition, PR_INTERVAL_NO_TIMEOUT);
		// Yay! Action
		switch (BlenderAction) {
			case BA_NEWBLENDFILE:
				if (debug>1) fprintf(stdout, "%i got BA_NEWBLENDFILE command\n", fd);
				// no need for the mutex lock anymore
				PR_Unlock(BlenderActionMutex);
				send_BlendFile(args);
				break;
			case BA_SINGLESHOTRENDER:
				if (debug>1) fprintf(stdout, "%i got BA_SINGLESHOTRENDER command\n", fd);
				// no need for the mutex lock anymore
				PR_Unlock(BlenderActionMutex);
				if (debug) fprintf(stdout, "%i is ready for rendering\n", fd);
				SingleShotLoop(args);
				break;
			case BA_QUIT:
				if (debug>1) fprintf(stdout, "%i got BA_QUIT command\n", fd);
				// no need for the mutex lock anymore
				PR_Unlock(BlenderActionMutex);
				PRB_Msg(btrSock, "Quit");
				// wait for the 210 message. Reuse nread and line
				nread = PRB_ReadLine(btrSock, line, MAXLINE);
				if (nread > 4) {
					if (Bstrncasecmp(line, "210",3) == 0) {
						// all's well, close connection
						PRB_Close(btrSock);
					} else {
						// TODO error parsing ... close anyway
						PRB_Close(btrSock);
					}
				} else {
					// TODO error parsing ... close anyway
					PRB_Close(btrSock);
				}
				free (args);
				return;
				break;
			case BA_STOPALLDAEMONS:
				if (debug>1) fprintf(stdout, "%i got BA_STOPALLDAEMONS command\n", fd);
				// no need for the mutex lock anymore
				PR_Unlock(BlenderActionMutex);
				PRB_Msg(btrSock, "StopDaemon");
				// wait for the 210 message. Reuse nread and line
				nread = PRB_ReadLine(btrSock, line, MAXLINE);
				if (nread > 4) {
					if (Bstrncasecmp(line, "210",3) == 0) {
						// all's well, close connection
						PRB_Close(btrSock);
					} else {
						// TODO error parsing ... close anyway
						PRB_Close(btrSock);
					}
				} else {
					// TODO error parsing ... close anyway
					PRB_Close(btrSock);
				}
				free (args);
				return;
				break;
			default:
				// no need for the mutex lock anymore
				PR_Unlock(BlenderActionMutex);
				if (debug>1) fprintf(stdout, "%i got unknown command\n", fd);
				// error++
			}
			// safeguard
	}
	if (debug>0) fprintf(stdout, "%i closed connection to %s:%i\n", fd, args->server, args->port);
	PRB_Close(btrSock);

	free (args);
	return;
}

void send_BlendFile(struct threadArgsType *args) {
	PRFileDesc *sockfd = args->sock2;
	int mySize = (int) BlendFileSize;
	char line[MAXLINE];
	PRInt32 nread;

	PRB_Msg(sockfd, "NewBlendFile; Size %d", mySize);
	// acquire a lock on the main thread to prevent render from beginning before the file
	// is all the way there.
	PR_Lock(mainThreadMutex);


	// and send the blendfile
	PRB_Send(sockfd, BlendFile, BlendFileSize, 0, PR_INTERVAL_NO_TIMEOUT);

	// wait for OK.
	if ( (nread = PRB_ReadLine(sockfd, line, MAXLINE)) <= 0) {
		if (debug>0) fprintf(stderr, "ERROR %i closed connection\n", args->fd2);
		PR_Unlock(mainThreadMutex);
		return;
	}
	for ( ; ; ) {
		if (debug>1) fprintf(stdout, "%i Command = [%s][%d]\n", args->fd2, line, nread);
		if (Bstrncasecmp(line,"250",3) == 0) {
			if (debug>0) fprintf(stdout, "%i accepted our blendfile\n", args->fd2);
			//PR_Unlock(mainThreadMutex);
			break;
		} else {
			if (debug>0) fprintf(stdout, "%i ERROR unknown response from renderd\n", args->fd2);
			PR_Unlock(mainThreadMutex);
			return;
		}
	}
	PR_Unlock(mainThreadMutex);
}

int read_blenderfile(PRFileDesc *brokersockfd, char *BlendFileName) {
	FILE *blendfd;
	int x;

	if ((blendfd = fopen(BlendFileName, "rb")) == NULL) {
		PRB_Msg(brokersockfd, "505 Error reading %s", BlendFileName);
		return 1;
	}
	fseek(blendfd, 0, SEEK_END);
	BlendFileSize = ftell(blendfd);
	fseek(blendfd, 0, SEEK_SET);
	brokerJob.BlendFileSize = BlendFileSize;
	if ((configInfo->minBlendFileSize > BlendFileSize) || (BlendFileSize > configInfo->maxBlendFileSize)) {
		PRB_Msg(brokersockfd, "504 Unacceptable BlendFileSize");
		fclose(blendfd);
		return 1;
	}
	if ((BlendFile = malloc(BlendFileSize)) == NULL) {
		PRB_Msg(brokersockfd, "504 BlendFile malloc problems");
		fclose(blendfd);
		return 1;
	}
	if ((x=fread(BlendFile, 1, BlendFileSize, blendfd)) != BlendFileSize) {
		PRB_Msg(brokersockfd, "504 BlendFile reading problems %d of %d",
		x, BlendFileSize);
		fclose(blendfd);
		return 1;
	}
	fclose(blendfd);
	return 0;
}

void stopdaemon(PRFileDesc *sockfd) {
	char dummyline[MAXLINE];

	PRB_Msg(sockfd, "210 Render Broker will exit, please disconnect");

	// prepare main thread exit
	StopDaemon = TRUE;

	// disconnect from all renderd's
	SignalThreads(BA_QUIT);

	// TODO wait for them to exit
	PR_Sleep(PR_SecondsToInterval(1));

	// wait for client to close connection (= failing readline)
	while (PRB_ReadLine(sockfd, dummyline, MAXLINE) >= 0);

	// close controller socket
	// note that this will generate a TIME_WAIT state for the socket
	// PRB_Close(sockfd);

	// close global listening socket
	PRB_Close(listenSock);
}

void stopalldaemons(PRFileDesc *sockfd) {
	char dummyline[MAXLINE];

	PRB_Msg(sockfd, "210 Render Broker will exit all daemons, please disconnect");

	// prepare main thread exit
	StopDaemon = TRUE;

	// disconnect from all renderd's
	SignalThreads(BA_STOPALLDAEMONS);

	// TODO wait for them to exit
	PR_Sleep(PR_SecondsToInterval(1));

	// wait for client to close connection (= failing readline)
	while (PRB_ReadLine(sockfd, dummyline, MAXLINE) >= 0);

	// close controller socket
	// note that this will generate a TIME_WAIT state for the socket
	// PRB_Close(sockfd);

	// close global listening socket
	PRB_Close(listenSock);
}

void showstatus(struct threadArgsType *args) {
	PRFileDesc *sockfd = args->sock;
	int i, total = 0;
	char timestring[MAXLINE];
	LONGINT uptime;
	char uptimeString[MAXLINE];
	PRExplodedTime exploded;
	struct SingleShotTag myshot;
	int readypct;
	char JobtimeString[MAXLINE];
	char *jstatus;

	PR_Lock(SingleShotMutex);
	// save current SingleShot struct
	myshot = SingleShot;
	PR_Unlock(SingleShotMutex);
	if (myshot.end == 0) {
		readypct = 100;
	} else {
		readypct = ((100 * (myshot.frame - 1)) / myshot.end);
	}

	PR_Lock(InfoMutex);
	PR_ExplodeTime(mainThread.epoch, PR_GMTParameters, &exploded);
	PRB_sprintExplodedTime(timestring, &exploded);
	uptime = (LONGINT) (PR_Now() - mainThread.epoch)/PR_USEC_PER_SEC;
	PRB_SecondsToString(uptimeString, uptime);
	PR_ExplodeTime(brokerJob.epoch, PR_GMTParameters, &exploded);
    PRB_sprintExplodedTime(JobtimeString, &exploded);

	for (i=0; i<MAXTHREADS; i++) {
		if (Info[i].occupied) total++;
	}
	if (brokerJob.status == B_INITIALIZE) {
		jstatus = "Initialize"; }
	else if (brokerJob.status == B_CURRENT) {
		jstatus = "Current";
	} else {
		jstatus = "Complete";
	}

	PRB_Msg(sockfd, "210 Blender broker %s up since %s (%s)",
	version, timestring, uptimeString);
	PRB_Msg(sockfd, "210 %3i%% ready. %i connections. %i images rendered at %.3f sec/frame",
	readypct, total, mainThread.frames_rendered,
	mainThread.msecPerFrame/1000.);

	for (i=0; i<MAXTHREADS; i++) {
		if (Info[i].occupied) {
			PRB_Msg(sockfd, "210 rendered %i frames at %.3f sec/frame at renderd %i %s:%i",
			Info[i].frames_rendered,
			Info[i].msecPerFrame / 1000.,
			i,
			Info[i].ip,
			Info[i].port);
		}
	}
	PR_Unlock(InfoMutex);
	PRB_Msg(sockfd, "\n210- Current Job Status:\n");
	PRB_Msg(sockfd, "210- Status             %s", jstatus);
	PRB_Msg(sockfd, "210- Filename           %s", brokerJob.BlendFileName);
	PRB_Msg(sockfd, "210- Filesize           %li", brokerJob.BlendFileSize);
	//TODO: brokerjob.scene prints out garbage after a file has been sent to renderds.
	//PRB_Msg(sockfd, "210- Scene              %s", brokerJob.scene);
	PRB_Msg(sockfd, "210- Start Frame        %i", brokerJob.start);
	PRB_Msg(sockfd, "210- End Frame          %i", brokerJob.end);
	PRB_Msg(sockfd, "210- Start Time         %s", JobtimeString);
	PRB_Msg(sockfd, "210- Current Frame      %i\n", SingleShot.frame);
}

void print_help()
{
	printf("Blender Network Renderer Broker options:\n");
	printf("  -n size\tSets the minimum bytes recognized as a valid .blend file\n");
	printf("  -x size\tSets the maximum bytes recognized as a valid .blend file\n");
	printf("  -i path\tSets the input directory to search for new files to render\n");
	printf("\t\tThis must be a fully qualified path to a write-enabled dir\n");
	//		printf("  -o path\tSets the output directory for finished renderings\n");
	//		printf("\t\t\tThis must be a fully-qualified path to a write-enabled dir\n");
	printf("  -p port\tSets the port number for the broker to accept renderd connections\n");
	printf("  -m port\tSets the minimum port number to spawn renderd threads on\n");
	printf("  -a port\tSets the maximum port number to spawn renderd threads on\n");
	printf("  -d level\tSets the debug level (default = 1)\n");
	printf("  -h\t\tPrints this helpful message!\n");
}

static void blender_esc(int sig)
{
	static int count = 0;

	if (sig == 2) {
		if (count) {
			SignalThreads(BA_STOPALLDAEMONS);
			printf("\nBroker killed\n");
			exit(2);
		}
		printf("\nSending stop daemon to renderd\n\n");
		SignalThreads(BA_STOPALLDAEMONS);
		printf("\nSent an internal break event. Press ^C again to kill Broker\n\n");
		count++;
	}
}

void brokerFileCopy (char *brokerCopyFrom, char *brokerCopyTo)
{
	PRDir *fd_dir;
	PRDirEntry *dirEntry;
	char filename[MAXLINE], newfilename[MAXLINE];

	// Make sure brokerwork and brokerdone exists
	if ((PR_MkDir(configInfo->brokerwork, 755) == PR_FAILURE) && (PR_GetError() != -5943)) {
		if (debug>1) fprintf(stderr, "broker_control: Cannot create %s, %s\n", configInfo->brokerwork, PRB_GetErrorText());
		PR_Delete(filename);
	}

	if ((PR_MkDir(configInfo->brokerdone, 755) == PR_FAILURE) && (PR_GetError() != -5943)) {
		if (debug>1) fprintf(stderr, "broker_control: Cannot create %s, %s\n", configInfo->brokerdone, PRB_GetErrorText());
		PR_Delete(filename);
	}

	if ((fd_dir = PR_OpenDir(brokerCopyFrom)) == NULL) {
		if (debug>0) fprintf(stdout, "broker_control:  %s not found.\n", brokerCopyFrom);
	} else {
		while ((dirEntry = PR_ReadDir(fd_dir, PR_SKIP_BOTH)) != NULL) {
			if (strstr(dirEntry->name, ".blend")) {
				if (debug>0) fprintf(stdout, "broker_control: found %s\n", dirEntry->name);
				strcpy(filename, brokerCopyFrom);
				strcat(filename, "/");
				strcat(filename, dirEntry->name);

				// now move blendfile away
				strcpy(newfilename, brokerCopyTo);
				strcat(newfilename, "/");
				strcat(newfilename, dirEntry->name);
				PR_Delete(newfilename);
				if (PR_SUCCESS != PR_Rename(filename,newfilename)) {
					if (debug>0) fprintf(stderr, "broker_control: move %s to %s failed: %s\n", filename, newfilename, PRB_GetErrorText());
				}
			}
			break;
		}
		PR_CloseDir(fd_dir);
	}
}

void brokerScan(char *brokerScanDir)
{
	PRDir *fd_dir;
	PRDirEntry *dirEntry;

	if (brokerWorkDone == B_WORK) {
		// Sleep 5 seconds for New Renderd to Broker initialization to occur.
		PR_Sleep(PR_SecondsToInterval(3));
	}

	// Check if the scandirectory exists, if it doesn't, create it
	if ((fd_dir = PR_OpenDir(brokerScanDir)) == NULL) {
		if (debug>0) fprintf(stdout, "brokerScan: scandirectory \"%s\" not found, trying to create...", brokerScanDir);
		
		if ((PR_MkDir(brokerScanDir, 755) == PR_FAILURE) && (PR_GetError() != -5943)) {
			if (debug>0) fprintf(stdout, "Error\n");
		}else{
			if (debug>0) fprintf(stdout, "Success\n");
		}
	}

	if ((fd_dir = PR_OpenDir(brokerScanDir)) == NULL) {				
	} else {
		while ((dirEntry = PR_ReadDir(fd_dir, PR_SKIP_BOTH)) != NULL) {
			if (strstr(dirEntry->name, ".blend")) {
				int startfr = 0, endfr = 0;
				char scenename[33];
				char filename[MAXLINE];
				char jobscene[33];
				scenename[0] = 0;

				// scan blendfile for frame numbers
				if (debug>0) fprintf(stdout, "broker_control: found %s\n", dirEntry->name);
				strcpy(filename, brokerScanDir);
				strcat(filename, "/");
				strcat(filename, dirEntry->name);

				scan_blendfile(filename, &startfr, &endfr, scenename);
				if (debug>0) fprintf(stdout, "filename is %s\n", filename);
				if (read_blenderfile(NULL, filename) != 0) {
					if (debug>1) fprintf(stdout, "ERROR in read_blenderfile\n");
				} else {
					if (debug>1) fprintf(stdout, "   read_blenderfile is OK\n");

					// file read ok, send action to threads
					SignalThreads(BA_NEWBLENDFILE);
					// send frame info to new renderd
					if (brokerWorkDone == B_WORK) {
						if (debug>0) fprintf(stdout, "broker_control: New Renderd frames %d-%d\n", SingleShot.frame, endfr);
					} else {
						if (debug>0) fprintf(stdout, "broker_control: Frames %d-%d for Scene %s\n", startfr, endfr, scenename);
						// Update BrokerJob stats
						strcpy (jobscene, scenename);
						brokerJob.status = B_INITIALIZE;
						brokerJob.BlendFileName = filename;
						brokerJob.start = startfr;
						brokerJob.end = endfr;
						brokerJob.scene = jobscene; 
						brokerJob.epoch = PR_Now();
					}
					// Windows refuses to issue rendering command without PR_Sleep here
					PR_Sleep(PR_SecondsToInterval(5));
					if (brokerWorkDone == B_WORK) {
						SetFrameLimit(STARTLIMIT, SingleShot.frame);
					} else {
						SetFrameLimit(STARTLIMIT, startfr);
					}
					SetFrameLimit(ENDLIMIT, endfr);
				}
				break;
			}
		}
	}
	PR_CloseDir(fd_dir);
}

int receive_BlendFile(PRFileDesc *sockfd)
{
	LONGINT BlendFileSize;
	FILE *blendfd;		
	void *BlendFile;
	PRInt32 nread;
	char line[MAXLINE],blendFileName[MAXLINE];
	char taskfile[MAXLINE];
	
	PR_Lock(mainThreadMutex);

	/* Read the name of the file */
	if (debug>1) fprintf(stdout, "Reading name..");
	if ( (nread = PRB_ReadLine(sockfd, blendFileName, MAXLINE)) == 0){
		PRB_Msg(sockfd, "Error while reading blendfilename");
		if (debug>1) fprintf(stdout, "ERROR\n");
		PR_Unlock(mainThreadMutex);
		return 0;
	}else{
		if (debug>1) fprintf(stdout, "SUCCESS\n");
	}
	
	/* Read the size of the file */
	if (debug>1) fprintf(stdout, "Reading filesize..");
	if ( (nread = PRB_ReadLine(sockfd, line, MAXLINE)) == 0){
		PRB_Msg(sockfd, "Error while reading blendfilesize");
		if (debug>1) fprintf(stdout, "ERROR\n");
		PR_Unlock(mainThreadMutex);
		return 0;
	}else{
		if (debug>1) fprintf(stdout, "SUCCESS (size: %s)\n", line);
		BlendFileSize = atol(line);		
	}	

	/* Alocate space for file */
	if ( (BlendFile = malloc(BlendFileSize)) == NULL ) {
		PRB_Msg(sockfd, "500 Not enough memory");
		PR_Unlock(mainThreadMutex);
		return 0;
	}	
	
	/* We're going to store the file at brokertasks/blendfilename */
	strcpy(taskfile, configInfo->brokertasks);
	strcat(taskfile, "/");
	strcat(taskfile, blendFileName);

	if (PRB_Recv(sockfd, BlendFile, BlendFileSize, 0, PR_INTERVAL_NO_TIMEOUT) == BlendFileSize) {
		if ((blendfd = fopen(taskfile, "wb")) == NULL) {
			PRB_Msg(sockfd, "500 Blender file fopen error");
			PR_Unlock(mainThreadMutex);
			return 0;
		} else {
			if (fwrite(BlendFile, BlendFileSize, 1, blendfd) != 1) {
				PRB_Msg(sockfd, "500 Blender file fwrite error");
				PR_Unlock(mainThreadMutex);
				return 0;
			} else {
				PRB_Msg(sockfd, "250 Blender file accepted");				
			}
			fclose(blendfd);
		}
	} else {
		PRB_Msg(sockfd, "500 Blender file not received properly");
		PR_Unlock(mainThreadMutex);
		return 0;
	}
	free(BlendFile);	
	
	if (debug>1) fprintf(stdout, "Exiting receive_BlendFile.\n");

	PR_Unlock(mainThreadMutex);
	return 1;
}