/* nspr-additions.c
 * Copyright (c) 2000 by Not a Number B.V. (NaN) and Hans Lambermont.
 * This file is covered by the license in the accompanying "LICENSE" file.
 * $Id: nspr-additions.c 74 2004-05-02 20:25:15Z goofster $
 */

#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include "nspr.h"
// for PR_FileDesc2NativeHandle:
#include "private/pprio.h"

#include "nspr-additions.h"
#include "common.h"

// these two should be mutex protected
extern int debug;
extern int StopDaemon;

// Our extensions to the PR_ library will use PRB_ (B from Blender)

PRInt32 PRB_Send(PRFileDesc *fd, void *buf, PRInt32 amount, PRIntn
		 flags, PRIntervalTime timeout) {
	PRInt32 nbytes;
	PRInt32 ntotal;
	char *bufPtr;

	bufPtr = buf;
	ntotal = 0;
	while (ntotal < amount) {
	nbytes = PR_Send(fd, bufPtr, amount-ntotal, flags, timeout);
	ntotal += nbytes;
	bufPtr += nbytes;
		if (debug>2) fprintf(stdout, "%6d bytes out (%6d to go) of the total %d\n",
				ntotal, amount - ntotal, amount);
	if (0 == nbytes) {
		fprintf(stderr, "PRB_Send failed because of a closed connection\n");
		return(0); // keep using NSPR error codes
	}
	if (-1 == nbytes) {
		fprintf(stderr, "PRB_Send failed because of error %s\n",
			PRB_GetErrorText());
		return(-1); // keep using NSPR error codes
	}
	}
	return(amount);
}

void PRB_Msg(PRFileDesc *fd, const char *fmt, ...) {
	va_list args;
	char txtbuf[MAXLINE];
	int txtbuflen;

	va_start(args, fmt);
	vsprintf(txtbuf, fmt, args);
	va_end(args);

	if (NULL == fd) {
		fprintf(stdout, " >  %s\n", txtbuf);
		} else {
		txtbuflen = strlen(txtbuf);
		if (debug>1) fprintf(stdout, "%d> %s\n", PRB_GetFileDesc(fd), txtbuf);
		PRB_Send(fd, strcat(txtbuf,"\r\n"), 2+txtbuflen,
			0, PR_INTERVAL_NO_TIMEOUT);
			}
	}

PRInt32 PRB_Recv(PRFileDesc *fd, void *buf, PRInt32 amount, PRIntn
				 flags, PRIntervalTime timeout) {
	PRInt32 nbytes;
	PRInt32 ntotal;
	char *bufPtr;
	char *receiveall;

	bufPtr = buf;
	ntotal = 0;
	receiveall = "Received data";

	while (ntotal < amount) {
		nbytes = PR_Recv(fd, bufPtr, amount-ntotal, flags, timeout);
		ntotal += nbytes;
		bufPtr += nbytes;
		if (amount > 1) {
			if (debug>1) fprintf(stdout,
				"%6d bytes in (%6d to go) of the total %d\n",
				ntotal, amount - ntotal, amount);
			}

		if (0 == nbytes) {
			if (!StopDaemon) fprintf(stderr,
				"PRB_Recv failed because of a closed connection\n");
			return(0); // keep using NSPR error codes
			}
		if (-1 == nbytes) {
			fprintf(stderr,
				"PRB_Recv failed because of error %s.\n",
				PRB_GetErrorText());
			return(-1); // keep using NSPR error codes
			}
		}
	return(amount);

		}

/* stupid and slow but threadsafe ReadLine function
 * PRB_ReadLine is now a macro case of PRB_ReadLineTime
 */
PRInt32 PRB_ReadLineTime(PRFileDesc *fd, char *buf, PRUint32 bufSize,
			PRIntervalTime timeout) {
	char *bufPtr = buf;
	int endFound;

	if (bufPtr == NULL) {
	fprintf(stderr, "PRB_ReadLine NULL pointer\n");
		return(-3);
	}
	*bufPtr =  0; // clear the buffer
	for (endFound = 0; !endFound; bufPtr++) {
	if (PRB_Recv(fd, bufPtr, 1, 0, timeout) == 1) {
		if (*bufPtr == '\r') {
		*bufPtr = 0;
		} else if (*bufPtr == '\n') {
		*bufPtr = 0;
		endFound = 1;
		}
	} else {
		if (!StopDaemon)
		fprintf(stderr, "PRB_ReadLine couldn't read 1 character\n");
		return(-1); // treat everything as an error
	}
	}
	return(bufPtr - buf);
}

void PRB_Close(PRFileDesc *fd) {
#ifdef WIN32
	int our_fd = PRB_GetFileDesc(fd);
	PR_Lock(mainThreadMutex);
	mainThread.nsprfd[our_fd] = 0;
	PR_Unlock(mainThreadMutex);
#endif
	if (PR_SUCCESS != PR_Close(fd)) {
	fprintf(stderr, "PR_Close error %s\n", PRB_GetErrorText());
	}
}

int PRB_GetFileDesc(PRFileDesc *fd) {
	int our_fd = 0;
#ifdef WIN32
	int i;
	PR_Lock(mainThreadMutex);
	for (i=1; i<MAXTHREADS; i++) {
	if (mainThread.nsprfd[i] == fd) {
		our_fd = i;
	}
	}
	if (our_fd == 0) {
	// not found, make a new one
	for (i=1; (i<MAXTHREADS) && (our_fd == 0); i++) {
		if (mainThread.nsprfd[i] == 0) {
		mainThread.nsprfd[i] = fd;
		our_fd = i;
		}
	}
	}
	PR_Unlock(mainThreadMutex);
#else
	our_fd = PR_FileDesc2NativeHandle(fd);
#endif
	return our_fd;
}

void PRB_sprintExplodedTime(char *string, const PRExplodedTime *et) {
	char *dayOfWeek[] =
	{ "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "???" };
	char *month[] =
	{ "Jan", "Feb", "Mar", "Apr", "May", "Jun",
	  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec", "???" };
	char tmpstring[MAXLINE];
	PRInt32 totalOffset;
	PRInt32 hourOffset, minOffset;
	char *sign;

	/* Print day of the week, month, day, hour, minute, and second */
	sprintf(string, "%s %s %ld %02ld:%02ld:%02ld ",
		dayOfWeek[et->tm_wday], month[et->tm_month], (LONGINT) et->tm_mday,
		(LONGINT) et->tm_hour, (LONGINT) et->tm_min, (LONGINT) et->tm_sec);

	/* Print time zone */
	totalOffset = et->tm_params.tp_gmt_offset + et->tm_params.tp_dst_offset;
	if (totalOffset == 0) {
	strcat(string, "UTC ");
	} else {
		sign = "";
		if (totalOffset < 0) {
		totalOffset = -totalOffset;
		sign = "-";
		}
		hourOffset = totalOffset / 3600;
		minOffset = (totalOffset % 3600) / 60;
	sprintf(tmpstring, "%s%02ld%02ld ", sign, (LONGINT) hourOffset,
		(LONGINT) minOffset);
		strcat(string, tmpstring);
	}

	/* Print year */
	sprintf(tmpstring, "%d", et->tm_year);
	strcat(string, tmpstring);
}

void PRB_SecondsToString(char *uptimeString, LONGINT seconds) {
	ldiv_t qr;
	LONGINT minutes, hours, days;
	if (seconds >= 60) {
	qr = ldiv(seconds, 60);
		minutes = qr.quot;
		seconds = qr.rem;
		if (minutes >= 60) {
		qr = ldiv(minutes, 60);
			hours = qr.quot;
			minutes = qr.rem;
			if (hours >= 24) {
		qr = ldiv(hours, 24);
				days = qr.quot;
				hours = qr.rem;
				sprintf(uptimeString, "%ld days %ld hours", days, hours);
			} else {
				sprintf(uptimeString, "%02ld:%02ld:%02ld", hours,
						minutes, seconds);
			}
		} else {
			sprintf(uptimeString, "00:%02ld:%02ld", minutes, seconds);
		}
	} else {
		sprintf(uptimeString, "00:00:%02ld", seconds);
	}
}

struct errinfo {
	PRErrorCode errcode;
	char *errname;
};

struct errinfo errcodes[] = {
{PR_OUT_OF_MEMORY_ERROR,		"PR_OUT_OF_MEMORY_ERROR"},
{PR_BAD_DESCRIPTOR_ERROR,		"PR_BAD_DESCRIPTOR_ERROR"},
{PR_WOULD_BLOCK_ERROR,			"PR_WOULD_BLOCK_ERROR"},
{PR_ACCESS_FAULT_ERROR, 		"PR_ACCESS_FAULT_ERROR"},
{PR_INVALID_METHOD_ERROR,		"PR_INVALID_METHOD_ERROR"},
{PR_ILLEGAL_ACCESS_ERROR,		"PR_ILLEGAL_ACCESS_ERROR"},
{PR_UNKNOWN_ERROR,				"PR_UNKNOWN_ERROR"},
{PR_PENDING_INTERRUPT_ERROR,	"PR_PENDING_INTERRUPT_ERROR"},
{PR_NOT_IMPLEMENTED_ERROR,		"PR_NOT_IMPLEMENTED_ERROR"},
{PR_IO_ERROR,					"PR_IO_ERROR"},
{PR_IO_TIMEOUT_ERROR,			"PR_IO_TIMEOUT_ERROR"},
{PR_IO_PENDING_ERROR,			"PR_IO_PENDING_ERROR"},
{PR_DIRECTORY_OPEN_ERROR,		"PR_DIRECTORY_OPEN_ERROR"},
{PR_INVALID_ARGUMENT_ERROR, 	"PR_INVALID_ARGUMENT_ERROR"},
{PR_ADDRESS_NOT_AVAILABLE_ERROR,"PR_ADDRESS_NOT_AVAILABLE_ERROR"},
{PR_ADDRESS_NOT_SUPPORTED_ERROR,"PR_ADDRESS_NOT_SUPPORTED_ERROR"},
{PR_IS_CONNECTED_ERROR, 		"PR_IS_CONNECTED_ERROR"},
{PR_BAD_ADDRESS_ERROR,			"PR_BAD_ADDRESS_ERROR"},
{PR_ADDRESS_IN_USE_ERROR,		"PR_ADDRESS_IN_USE_ERROR"},
{PR_CONNECT_REFUSED_ERROR,		"PR_CONNECT_REFUSED_ERROR"},
{PR_NETWORK_UNREACHABLE_ERROR,	"PR_NETWORK_UNREACHABLE_ERROR"},
{PR_CONNECT_TIMEOUT_ERROR,		"PR_CONNECT_TIMEOUT_ERROR"},
{PR_NOT_CONNECTED_ERROR,		"PR_NOT_CONNECTED_ERROR"},
{PR_LOAD_LIBRARY_ERROR, 		"PR_LOAD_LIBRARY_ERROR"},
{PR_UNLOAD_LIBRARY_ERROR,		"PR_UNLOAD_LIBRARY_ERROR"},
{PR_FIND_SYMBOL_ERROR,			"PR_FIND_SYMBOL_ERROR"},
{PR_INSUFFICIENT_RESOURCES_ERROR,"PR_INSUFFICIENT_RESOURCES_ERROR"},
{PR_DIRECTORY_LOOKUP_ERROR, 	"PR_DIRECTORY_LOOKUP_ERROR"},
{PR_TPD_RANGE_ERROR,			"PR_TPD_RANGE_ERROR"},
{PR_PROC_DESC_TABLE_FULL_ERROR, "PR_PROC_DESC_TABLE_FULL_ERROR"},
{PR_SYS_DESC_TABLE_FULL_ERROR,	"PR_SYS_DESC_TABLE_FULL_ERROR"},
{PR_NOT_SOCKET_ERROR,			"PR_NOT_SOCKET_ERROR"},
{PR_NOT_TCP_SOCKET_ERROR,		"PR_NOT_TCP_SOCKET_ERROR"},
{PR_SOCKET_ADDRESS_IS_BOUND_ERROR,	"PR_SOCKET_ADDRESS_IS_BOUND_ERROR"},
{PR_NO_ACCESS_RIGHTS_ERROR, 	"PR_NO_ACCESS_RIGHTS_ERROR"},
{PR_OPERATION_NOT_SUPPORTED_ERROR,"PR_OPERATION_NOT_SUPPORTED_ERROR"},
{PR_PROTOCOL_NOT_SUPPORTED_ERROR,"PR_PROTOCOL_NOT_SUPPORTED_ERROR"},
{PR_REMOTE_FILE_ERROR,			"PR_REMOTE_FILE_ERROR"},
{PR_BUFFER_OVERFLOW_ERROR,		"PR_BUFFER_OVERFLOW_ERROR"},
{PR_CONNECT_RESET_ERROR,		"PR_CONNECT_RESET_ERROR"},
{PR_RANGE_ERROR,				"PR_RANGE_ERROR"},
{PR_DEADLOCK_ERROR, 			"PR_DEADLOCK_ERROR"},
{PR_FILE_IS_LOCKED_ERROR,		"PR_FILE_IS_LOCKED_ERROR"},
{PR_FILE_TOO_BIG_ERROR, 		"PR_FILE_TOO_BIG_ERROR"},
{PR_NO_DEVICE_SPACE_ERROR,		"PR_NO_DEVICE_SPACE_ERROR"},
{PR_PIPE_ERROR, 				"PR_PIPE_ERROR"},
{PR_NO_SEEK_DEVICE_ERROR,		"PR_NO_SEEK_DEVICE_ERROR"},
{PR_IS_DIRECTORY_ERROR, 		"PR_IS_DIRECTORY_ERROR"},
{PR_LOOP_ERROR, 				"PR_LOOP_ERROR"},
{PR_NAME_TOO_LONG_ERROR,		"PR_NAME_TOO_LONG_ERROR"},
{PR_FILE_NOT_FOUND_ERROR,		"PR_FILE_NOT_FOUND_ERROR"},
{PR_NOT_DIRECTORY_ERROR,		"PR_NOT_DIRECTORY_ERROR"},
{PR_READ_ONLY_FILESYSTEM_ERROR, "PR_READ_ONLY_FILESYSTEM_ERROR"},
{PR_DIRECTORY_NOT_EMPTY_ERROR,	"PR_DIRECTORY_NOT_EMPTY_ERROR"},
{PR_FILESYSTEM_MOUNTED_ERROR,	"PR_FILESYSTEM_MOUNTED_ERROR"},
{PR_NOT_SAME_DEVICE_ERROR,		"PR_NOT_SAME_DEVICE_ERROR"},
{PR_DIRECTORY_CORRUPTED_ERROR,	"PR_DIRECTORY_CORRUPTED_ERROR"},
{PR_FILE_EXISTS_ERROR,			"PR_FILE_EXISTS_ERROR"},
{PR_MAX_DIRECTORY_ENTRIES_ERROR,"PR_MAX_DIRECTORY_ENTRIES_ERROR"},
{PR_INVALID_DEVICE_STATE_ERROR, "PR_INVALID_DEVICE_STATE_ERROR"},
{PR_DEVICE_IS_LOCKED_ERROR, 	"PR_DEVICE_IS_LOCKED_ERROR"},
{PR_NO_MORE_FILES_ERROR,		"PR_NO_MORE_FILES_ERROR"},
{PR_END_OF_FILE_ERROR,			"PR_END_OF_FILE_ERROR"},
{PR_FILE_SEEK_ERROR,			"PR_FILE_SEEK_ERROR"},
{PR_FILE_IS_BUSY_ERROR, 		"PR_FILE_IS_BUSY_ERROR"},
{PR_IN_PROGRESS_ERROR,			"PR_IN_PROGRESS_ERROR"},
{PR_ALREADY_INITIATED_ERROR,	"PR_ALREADY_INITIATED_ERROR"},
{PR_GROUP_EMPTY_ERROR,			"PR_GROUP_EMPTY_ERROR"},
{PR_INVALID_STATE_ERROR,		"PR_INVALID_STATE_ERROR"},
{PR_NETWORK_DOWN_ERROR, 		"PR_NETWORK_DOWN_ERROR"},
{PR_SOCKET_SHUTDOWN_ERROR,		"PR_SOCKET_SHUTDOWN_ERROR"},
{PR_CONNECT_ABORTED_ERROR,		"PR_CONNECT_ABORTED_ERROR"},
{PR_HOST_UNREACHABLE_ERROR, 	"PR_HOST_UNREACHABLE_ERROR"}
};

int numerrors = sizeof(errcodes)/sizeof(errcodes[0]);

char *PRB_ErrorText(PRErrorCode err) {
	int i;
	for (i=0; i<numerrors; i++)
	if (err == errcodes[i].errcode)
		return errcodes[i].errname;
	return "Undefined NSPR Error";
}

char *PRB_GetErrorText() {
   return PRB_ErrorText( PR_GetError() );
}
