#include <string.h>
#include <stdio.h>

#include "parser.h"
#include "compat.h"

#ifdef WIN32
  #define BLENDEREXEDEFAULT "blender.exe"
#else
  #define BLENDEREXEDEFAULT "./blender"
#endif

#define MAXLINELEN 256

static char *parse_next_token(FILE *in);
static void parse_print_info(ConfigInfo *ci);

// keep track of the current input line number
static int cur_line = 0;

// set our debug level from the calling program
extern int debug;

/*
 * setup the default values and parse the config file
 * if one exists
*/
ConfigInfo *parse_config(void) {
	ConfigInfo *ci = (ConfigInfo *)malloc(sizeof(ConfigInfo));
	FILE *in;

	// set the defaults
	ci->minBlendFileSize = 5;
	ci->maxBlendFileSize = 10000000;
	ci->renderdBlendFileName = strdup("renderd.blend");
	ci->blenderexe = strdup(BLENDEREXEDEFAULT);
	ci->brokertasks = strdup("brokertasks");
	ci->brokerdone = strdup("brokerdone");
	ci->brokerwork = strdup("brokerwork");
	ci->scansubnet = 1;
	ci->brokerip = strdup("0.0.0.0");
	ci->renderdPort = 5002;
	ci->brokerPort = 5001;
	ci->minPort = 1024;
	ci->maxPort = 6000;
	ci->debug = 1;

	// if there is no config file we just use the defaults
	if (NULL != (in = fopen(CONFIGFILE, "r"))) {
		while(!feof(in)) {
			char *token = parse_next_token(in);
			char *val = parse_next_token(in);

			// this is probably the eof condition, but
			// left the while break the loop
			if (strlen(token) == 0)
				continue;

			if (!Bstrncasecmp(token, "MINBLENDFILESIZE", 16)){
				ci->minBlendFileSize = atol(val);

			}else if (!Bstrncasecmp(token, "MAXBLENDFILESIZE", 16)){
				ci->maxBlendFileSize = atol(val);

			}else if (!Bstrncasecmp(token, "RENDERDBLENDFILENAME", 20)) {
				free(ci->renderdBlendFileName);
				ci->renderdBlendFileName = strdup(val);

			}else if (!Bstrncasecmp(token, "BLENDEREXE", 10)) {
				free(ci->blenderexe);
				ci->blenderexe = strdup(val);

			}else if (!Bstrncasecmp(token, "INPUTDIRECTORY", 14)) {
				free(ci->brokertasks);
				ci->brokertasks = strdup(val);

			}else if (!Bstrncasecmp(token, "DONEDIRECTORY", 13)) {
				free(ci->brokerdone);
				ci->brokerdone = strdup(val);

			}else if (!Bstrncasecmp(token, "WORKDIRECTORY", 13)) {
				free(ci->brokerwork);
				ci->brokerwork = strdup(val);

			}else if (!Bstrncasecmp(token, "SCANSUBNET", 10)){
				ci->scansubnet = atoi(val);

			}else if (!Bstrncasecmp(token, "BROKERIP", 8)) {
				free(ci->brokerip);
				ci->brokerip = strdup(val);

			}else if (!Bstrncasecmp(token, "RENDERDPORT", 11)){
				ci->renderdPort = atoi(val);

			}else if (!Bstrncasecmp(token, "BROKERPORT", 10)){
				ci->brokerPort = atoi(val);

			}else if (!Bstrncasecmp(token, "MINPORT", 7)){
				ci->minPort = atoi(val);

			}else if (!Bstrncasecmp(token, "MAXPORT", 7)){
				ci->maxPort = atoi(val);

			}else if (!Bstrncasecmp(token, "DEBUG", 5)){
				ci->debug = atoi(val);
			}
			else
				fprintf(stdout, "Unknown token (%s) on line (%d)\n", token, cur_line);

			free(val);
			free(token);
		}
		fclose(in);
	}

	// if you want to see the read in values for the config
	// info turn this on
	if (debug>1)
		parse_print_info(ci);

	return ci;
}

/*
 * Get the next token from the input stream
*/
char *parse_next_token(FILE *in) {
	char *token = (char *)malloc(sizeof(char) * MAXLINELEN);
	int curPos = 0;
	int comment = 0;
	int c;

	memset(token, 0, sizeof(char) * MAXLINELEN);
	while((c = getc(in)) != EOF) {
		unsigned char ch = (unsigned char)c;

		// on a comment skip to the end of the line
		if (comment && (ch != '\n'))
			continue;

		if (ch == '\n') {
			comment = 0;
			cur_line ++;

			// if we have input return it
			if (curPos > 0)
				break;

		} else if (ch == '#') {
			comment = 1;

		} else if ((ch == ' ') || (ch == '\t')) {
			// if we have input then return that
			if (curPos > 0)
				break;

		} else {
			// a normal char, append to input buffer
			token[curPos] = c;
			curPos++;

			// this is -1 so we can have a \0 on the end of the line
			if (curPos == (MAXLINELEN - 1)) {
				fprintf(stdout, "Error: Line (%d) to long in config file\n", cur_line);
				exit(-1);
			}
		}
	}
	return token;
}

static void parse_print_info(ConfigInfo *ci) {
	fprintf(stdout, "Option Values after Config File:\n");
	fprintf(stdout, "  Min blend file size: %f\n", (float)ci->minBlendFileSize);
	fprintf(stdout, "  Max blend file size: %f\n", (float)ci->maxBlendFileSize);
	fprintf(stdout, "  Renderd blend file name: %s\n", ci->renderdBlendFileName);
	fprintf(stdout, "  Blender executable: %s\n", ci->blenderexe);
	fprintf(stdout, "  Input Directory: %s\n", ci->brokertasks);
	fprintf(stdout, "  Work Directory: %s\n", ci->brokerwork);
	fprintf(stdout, "  Output Directory: %s\n", ci->brokerdone);
	fprintf(stdout, "  Renderd port: %d\n", ci->renderdPort);
	fprintf(stdout, "  Broker port: %d\n", ci->brokerPort);
	fprintf(stdout, "  Min port: %d\n", ci->minPort);
	fprintf(stdout, "  Max port: %d\n", ci->maxPort);
	fprintf(stdout, "  Debug: %i\n", ci->debug);
}
