/* renderd.c
 * Copyright (c) 2000 by Not a Number B.V. (NaN) and Hans Lambermont.
 * This file is covered by the license in the accompanying "LICENSE" file.
 * $Id: renderd.c 83 2004-10-24 05:48:40Z mcbridea $
 */

// system includes
#include <string.h>
#include <stdio.h>
#include "plgetopt.h"
#include "nspr.h"

// our own includes
#include "nspr-additions.h"
#include "compat.h"
#include "scan.h"
#include "common.h"
#include "parser.h"
#include "renderd.h"

int main(int argc, char **argv)
{
	PRFileDesc *clientSock;
	PRNetAddr listenAddr, clientAddr;
	int i;
	char *BlenderEnv;
	PLOptStatus os;
	PLOptState *opt;

	PR_STDIO_INIT();

	// Print program name and version
	fprintf(stdout, "Blender Renderdaemon Renderd\n");
	fprintf(stdout, "Version: %s\n\n", version);

	// Parse config file if it exists
	configInfo = parse_config();

	// Set Debug from configfile
	debug = configInfo->debug;

	// Read and execute any command line arguments

	opt = PL_CreateOptState(argc, argv, "b:d:f:hn:p:r:s:");
	while (PL_OPT_EOL != (os = PL_GetNextOpt(opt)))
	{
		if (PL_OPT_BAD == os) continue;

		switch (opt->option)
		{
			case 0:
				if (debug>1)
				{
					fprintf(stdout, "option %c", opt->option);
					if (opt->value)
						fprintf(stdout, " with arg %s", opt->value);
					fprintf(stdout, "\n");
				}
				break;
			case 'b':
				if (debug>1) fprintf(stdout, "blender set to %s\n", opt->value);
				free(configInfo->blenderexe);
				configInfo->blenderexe = strdup(opt->value);
				break;
			case 'd':
				if (debug>1) fprintf(stdout, "debug set to %s\n", opt->value);
				debug = atoi(opt->value);
				break;
			case 'f':
				if (debug>1) fprintf(stdout, "blender file name set to %s\n", opt->value);
				free(configInfo->renderdBlendFileName);
				configInfo->renderdBlendFileName = strdup(opt->value);
				break;
			case 'h':
				print_help();
				exit(0);
			case 'n':
				if (debug>1) fprintf(stdout, "broker address set to %s\n", opt->value);
//				configInfo->brokerPort = atoi(opt->value); Not yet sure how to specify a broker IP
				break;
			case 'p':
				if (debug>1) fprintf(stdout, "renderd port set to %s\n", opt->value);
				configInfo->renderdPort = atoi(opt->value);
				break;
			case 'r':
				if (debug>1) fprintf(stdout, "broker port set to %s\n", opt->value);
				configInfo->brokerPort = atoi(opt->value);
				break;
			case 's':
				if (debug>1) fprintf(stdout, "max image file size set to %s\n", opt->value);
				configInfo->maxBlendFileSize = atoi(opt->value);
				break;
			default:
				print_help();
				exit(0);
		}
	}

	PL_DestroyOptState(opt);

	BlenderEnv = PR_GetEnv("blender");
	if ((NULL == BlenderEnv) || ('\0' == *BlenderEnv)) {
		strcpy(BLENDEREXE, configInfo->blenderexe);
	} else {
		strcpy(BLENDEREXE , BlenderEnv);
	}

	if (debug>0) fprintf(stdout, "Using BLENDER = %s\n", BLENDEREXE);

	InfoMutex = PR_NewLock();
	mainThreadMutex = PR_NewLock();

	PR_Lock(InfoMutex);
	for (i=0; i<MAXTHREADS; i++) {
		Info[i].occupied = FALSE;
		Info[i].frames_rendered = 0;
	}
	PR_Unlock(InfoMutex);

	PR_Lock(mainThreadMutex);
	mainThread.renderThreadfd = 0;
	mainThread.frames_rendered = 0;
	mainThread.epoch = PR_Now();
	mainThread.msecPerFrame = 0;
	for (i=0; i<MAXTHREADS; i++) {
		mainThread.nsprfd[i] = 0;
	}
	PR_Unlock(mainThreadMutex);

	port.local = configInfo->renderdPort;
	port.remote = configInfo->brokerPort;

	listenSock = PR_NewTCPSocket();
	memset(&listenAddr, 0 , sizeof(listenAddr));
	listenAddr.inet.family = PR_AF_INET;
	listenAddr.inet.port = PR_htons(port.local);
	listenAddr.inet.ip = PR_htonl(PR_INADDR_ANY);
	if (PR_Bind(listenSock, &listenAddr) == PR_FAILURE) {
		if(debug>0) fprintf(stderr, "Could not create socket. Error %s (just wait TIME_WAIT)\n", PRB_GetErrorText());
		exit(1);
	}
	if (PR_Listen(listenSock, 10) == PR_FAILURE) {
		if(debug>0) fprintf(stderr, "Unable to listen on port %i, error %s.\n", port.local, PRB_GetErrorText());
		exit(1);
	}

	// we might not want to scan the net. make this optional
	if(configInfo->scansubnet == 1){
		if ( ! PR_CreateThread(PR_USER_THREAD, renderd_scan_thread, (void *) &port, PR_PRIORITY_NORMAL, PR_GLOBAL_THREAD, PR_UNJOINABLE_THREAD, 0) ) {
			if(debug>0) fprintf(stderr, "Cannot create scan thread\n");
			exit(1);
		}
	}

	// main thread starting loop
	for ( ; ; ) {
		if ((clientSock = PR_Accept(listenSock, &clientAddr, FOREVER)) != NULL) {
			// Save clientAddr info, indexed by socket file descriptor
			struct threadArgsType *threadArgs =
			malloc(sizeof(struct threadArgsType));
			// use the real filedescriptor, good for debugging
			int fd = PRB_GetFileDesc(clientSock);
			if (fd < MAXTHREADS) {
				threadArgs->fd = fd;
				threadArgs->sock = clientSock;
				threadArgs->addr = clientAddr;
				if ( ! PR_CreateThread(PR_USER_THREAD, renderd_thread, threadArgs, PR_PRIORITY_NORMAL, PR_GLOBAL_THREAD, PR_UNJOINABLE_THREAD, 0)) {
					if (debug>1) fprintf(stderr, "Could not create dumb thread %s (%d)!\n", PRB_GetErrorText(), PR_GetOSError());
				}
			} else {
				PRB_Close(clientSock);
				if (debug>0) fprintf(stdout, "Too many threads (%d >= max %i)\n", fd, MAXTHREADS-1);
			}
		} else if (StopDaemon) break;
	}
	if (debug>0) fprintf(stdout, "Main thread exit\n");
	return(0);
}

void PR_CALLBACK renderd_thread(void *arg)
{
	struct threadArgsType *args = (struct threadArgsType*) arg;
	PRFileDesc *sockfd = args->sock;
	PRInt32 nread;
	char line[MAXLINE];
	char blendFileName[MAXLINE];
	int fd = args->fd;

	PR_Lock(InfoMutex);
	Info[fd].occupied = TRUE;
	Info[fd].epoch = PR_Now();
	PR_NetAddrToString(&args->addr, Info[fd].ip, 256);
	Info[fd].port = PR_ntohs(args->addr.inet.port);
	if (debug) fprintf(stdout, "%d got connection from %s:%u\n", fd, Info[fd].ip, Info[fd].port);
	PR_Unlock(InfoMutex);

	sprintf(blendFileName, "%s-%i",configInfo->renderdBlendFileName,fd);
	PRB_Msg(sockfd, "220 Blender renderd %s ready", version);
	for ( ; ; ) {
		if ( (nread = PRB_ReadLine(sockfd, line, MAXLINE)) == 0)
			break;		/* connection closed by other end */
		if (nread<0) {
			if (debug>1) fprintf(stdout, "renderd thread exits\n");
			break;
		}
		if (debug>1) fprintf(stdout, "%i< %s\n", fd, line);
		if (Bstrncasecmp(line,"Help",4) == 0) {
			showhelp(sockfd, line + 4);
		} else if (Bstrncasecmp(line,"Quit",4) == 0) {
			PRB_Msg(sockfd, "210 Render Deamon connection will be closed");
			PRB_Close(sockfd);
			break;
		} else if (Bstrncasecmp(line,"Status",6) == 0) {
			showstatus(sockfd);
		} else if (Bstrncasecmp(line,"Config",6) == 0) {
			showconfig(sockfd);
		} else if (Bstrncasecmp(line,"StopDaemon",10) == 0) {
			stopdaemon(sockfd);
			break;
		} else if (Bstrncasecmp(line,"NewBlendFile; Size ",19) == 0) {
			accept_blenderfile(sockfd, line + 19, blendFileName);
		} else if (Bstrncasecmp(line,"SingleShotRender",16) == 0) {
			singleshotrender(arg, line + 16, blendFileName);
		} else if (Bstrncasecmp(line,"Debug ", 6) == 0) {
			debug = atoi(line + 6);
			PRB_Msg(sockfd, "210 debug set to %i", debug);
		}
	}
	// clean the temporary blendFileName
	PR_Delete(blendFileName);
	free(args);
	PR_Lock(InfoMutex);
	Info[fd].occupied = FALSE;
	Info[fd].frames_rendered = 0;
	PR_Unlock(InfoMutex);
	if (debug>1) fprintf(stdout, "Exiting renderd_thread.\n");
}

void singleshotrender(struct threadArgsType *args, char *linep, char *blendFileName)
{
	int framenr;
	LONGINT framefilesize;
	char blendercommand[MAXLINE];
	char blenderresult[MAXLINE];
	char *saveptr;
	char framefilename[MAXLINE];
	FILE *blenderfd, *framefd;
	void *frame;
	PRFileDesc *sockfd = args->sock;
	PRIntervalTime startTime = PR_IntervalNow();
	PRUint32 timeDiff;

	framenr = atol(linep);
	if (debug) fprintf(stdout, "Blender ordered to render frame %i\n", framenr);
	sprintf(blendercommand, "%s -b %s -f %i", BLENDEREXE, blendFileName,
	framenr);
	if (debug>1) fprintf(stdout, "blendercommand = %s\n", blendercommand);
	if ((blenderfd = Bpopen(blendercommand, "r")) != NULL) {
		memset(&blenderresult, 0, MAXLINE);
		for ( ; ; ) {
		if (fread(&blenderresult, 1, MAXLINE, blenderfd) == 0) {
			PRB_Msg(sockfd, "503 blenderpipe [%s] failed",
			blendercommand);
			Bpclose(blenderfd);
			return;
		}
			if ((saveptr = strstr(blenderresult, "Saved:")) != NULL) {
		Bpclose(blenderfd);
				break;
			}
		}
	} else {
		PRB_Msg(sockfd, "502 blendercommand [%s] failed",
		blendercommand);
		return;
	}
	/* blender does not use errorcodes yet :-( so start digging through
	* it's text output */
	if (!strstr(blenderresult, "Saved:")) {
		PRB_Msg(sockfd, "504 blender render command failed");
		fprintf(stdout, "[%s]\n", blenderresult);
		return;
	} else {
		if (debug>2) fprintf(stdout, "[%s]\n", blenderresult);
	}
	if ((saveptr = strstr(blenderresult, "Saved:")) != NULL) {
		if (1 != sscanf(saveptr, "Saved: %s Time:", framefilename)) {
			// darn, format error
			if (debug>1) fprintf(stdout, "ERROR. No 'Saved: <filename> Time:' found in [%s]\n", blenderresult);
			if (debug>1) fprintf(stdout, "Output from: %s\n", blendercommand);
			return;
		}
	} else {
		if (debug>1) fprintf(stdout, "ERROR. No 'Saved:' found in [%s]\n", blenderresult);
		if (debug>1) fprintf(stdout, "Output from: %s\n", blendercommand);
		return;
	}

	/* immediately send back rendered frame */
	if ((framefd = fopen(framefilename, "rb")) == NULL) {
		PRB_Msg(sockfd, "505 Error reading %s", framefilename);
		return;
	}
	fseek(framefd, 0, SEEK_END);
	framefilesize = ftell(framefd);
	fseek(framefd, 0, SEEK_SET);
	if ((configInfo->minBlendFileSize > framefilesize) || (framefilesize > configInfo->maxBlendFileSize)) {
		PRB_Msg(sockfd, "504 Unacceptable framefilesize");
		return;
	}
	if ((frame = malloc(framefilesize)) == NULL) {
		PRB_Msg(sockfd, "504 Framefile malloc problems");
		return;
	}
	if ((LONGINT) fread(frame, 1, framefilesize, framefd) != framefilesize) {
		PRB_Msg(sockfd, "504 Framefile reading problems");
		return;
	}
	fclose(framefd);

	PRB_Msg(sockfd, "250 Prepare for frame %i %s upload of %ld bytes",
	framenr, framefilename, framefilesize);

	// finally put the frame on the wire
	if (PRB_Send(sockfd, frame, framefilesize, 0, PR_INTERVAL_NO_TIMEOUT)!=framefilesize){
		if (debug>0) fprintf(stderr, "Error sending frame back to broker!\n");
	}
	free(frame);
	if (PR_Delete(framefilename) != PR_SUCCESS){
		if (debug>0) fprintf(stderr, "Error deleting finished file: %s\n",PRB_GetErrorText());
	}

	if (debug>0) fprintf(stdout, "Frame %s uploaded to broker.\n", framefilename);

	// Update statistics
	timeDiff = PR_IntervalToMilliseconds(PR_IntervalNow() - startTime);
	PR_Lock(InfoMutex);
	// for this fd
	Info[args->fd].msecPerFrame = (
	Info[args->fd].msecPerFrame *
	Info[args->fd].frames_rendered + timeDiff) /
	++(Info[args->fd].frames_rendered);
	// and for the totals
	mainThread.msecPerFrame = (
	mainThread.msecPerFrame *
	mainThread.frames_rendered + timeDiff) /
	++(mainThread.frames_rendered);
	PR_Unlock(InfoMutex);
	if (debug>1) fprintf(stdout, "Exiting singleshotrender thread.\n");
}

void accept_blenderfile(PRFileDesc *sockfd, char *linep, char *blendFileName)
{
	LONGINT BlendFileSize;
	FILE *blendfd;
	BlendFileSize = atol(linep);

	if ((configInfo->minBlendFileSize > BlendFileSize) && (BlendFileSize > configInfo->maxBlendFileSize)) {
		PRB_Msg(sockfd, "501 Unacceptable BlendFileSize");
	} else {
		void *BlendFile;
		if ( (BlendFile = malloc(BlendFileSize)) == NULL ) {
			PRB_Msg(sockfd, "500 Not enough memory");
			return;
		}

	if (PRB_Recv(sockfd, BlendFile, BlendFileSize, 0, PR_INTERVAL_NO_TIMEOUT) == BlendFileSize) {
		if ((blendfd = fopen(blendFileName, "wb")) == NULL) {
			PRB_Msg(sockfd, "500 Blender file fopen error");
		} else {
			if (fwrite(BlendFile, BlendFileSize, 1, blendfd) != 1) {
				PRB_Msg(sockfd, "500 Blender file fwrite error");
			} else {
				PRB_Msg(sockfd, "250 Blender file accepted");
			}
			fclose(blendfd);
		}
	} else {
		PRB_Msg(sockfd, "500 Blender file not received properly");
	}
	free(BlendFile);
	//		} else {
	//		PRB_Msg(sockfd, "501 Unacceptable BlendFileSize");
	}
	if (debug>1) fprintf(stdout, "Exiting accept_blenderfile thread.\n");
}

void stopdaemon(PRFileDesc *sockfd)
{
	char dummyline[MAXLINE];

	PRB_Msg(sockfd, "210 Render Deamon will exit, please disconnect");

	// prepare main thread exit
	StopDaemon = TRUE;

	// wait for client to close connection (= failing readline)
	while (PRB_ReadLine(sockfd, dummyline, MAXLINE) >= 0);

	// close global listening socket
	PRB_Close(listenSock);
}

void showstatus(PRFileDesc *sockfd)
{
	int i, total = 0;
	char timestring[MAXLINE];
	LONGINT uptime;
	char uptimeString[MAXLINE];
	PRExplodedTime exploded;

	PR_Lock(InfoMutex);
	PR_ExplodeTime(mainThread.epoch, PR_GMTParameters, &exploded);
	PRB_sprintExplodedTime(timestring, &exploded);
	uptime = (LONGINT) (PR_Now() - mainThread.epoch)/PR_USEC_PER_SEC;
	PRB_SecondsToString(uptimeString, uptime);

	for (i=0; i<MAXTHREADS; i++) {
		if (Info[i].occupied) total++;
	}
	PRB_Msg(sockfd, "210 Blender renderd %s up since %s (%s)",
	version, timestring, uptimeString);
	PRB_Msg(sockfd, "210 %i connections. %i images rendered at %.1f sec/frame",
	total, mainThread.frames_rendered, mainThread.msecPerFrame / 1000.);
	for (i=0; i<MAXTHREADS; i++) {
		if (Info[i].occupied) {
			PRB_Msg(sockfd,
			"210 rendered %i frames at %.1f sec/frame for broker %i %s:%i",
			Info[i].frames_rendered,
			Info[i].msecPerFrame / 1000.,
			i,
			Info[i].ip,
			Info[i].port);
		}
	}
	PR_Unlock(InfoMutex);
}

void showconfig(PRFileDesc *sockfd)
{
	PRB_Msg(sockfd, "210-This is the Blender renderd renderdaemon");
	PRB_Msg(sockfd, "210-Current configuration settings:");
	PRB_Msg(sockfd, "210- Blender:\t\t%s",configInfo->blenderexe);
	PRB_Msg(sockfd, "210- Filename:\t\t%s",configInfo->renderdBlendFileName);
	PRB_Msg(sockfd, "210- Max Image Size:\t%i", configInfo->maxBlendFileSize);
//	PRB_Msg(sockfd, "210- Broker Address:\t%i",configInfo->brokerPort); Not sure how yet
	PRB_Msg(sockfd, "210- Broker Port:\t%i",configInfo->brokerPort);
	PRB_Msg(sockfd, "210- This RenderD Port:\t%i",configInfo->renderdPort);
	PRB_Msg(sockfd, "210- Debug:\t\t%i",debug);
}


void showhelp(PRFileDesc *sockfd, char *linep)
{
	PRB_Msg(sockfd, "210-This is the Blender renderdaemon");
	PRB_Msg(sockfd, "210-Available commands are:");
	PRB_Msg(sockfd, "210-Help                    this helpfile");
	PRB_Msg(sockfd, "210-Quit                    end connection");
	PRB_Msg(sockfd, "210-StopDaemon              make renderd exit. handle with care");
	PRB_Msg(sockfd, "210-Status                  give renderd and blender status");
	PRB_Msg(sockfd, "210-Config                  show current renderd config settings");
	PRB_Msg(sockfd, "210-Debug xxx               set debugging to level xxx");
	PRB_Msg(sockfd, "210-NewBlendFile; Size xxx  upload new blender file with size = xxx bytes");
	PRB_Msg(sockfd, "210-                        then start sending the file");
	PRB_Msg(sockfd, "210 SingleShotRender xxx    render frame xxx of current blenderfile");
}

void print_help()
{
	printf("Blender Network Renderer options:\n");
	printf("  -b path\tSets the exact path & name of the blender executable\n");
	printf("  -f filename\tSets the temp filename for the remote blender renderer\n");
	printf("\t\tThis will be saved in the same directory as blender.\n");
	printf("  -s size\tSets the maximum image file size to transmit\n");
	printf("  -p port\tSets the port number for renderd to accept broker connections\n");
	printf("  -n address\tSets the IP address of the broker to use\n");
	printf("  -r port\tSets the port to look for the broker on\n");
	printf("  -d level\tSets the debug level (default = 1)\n");
	printf("  -h\t\tPrints this helpful message!\n");
}

